// Listing9-5.S

    #include    "ao32aa.inc"

    .arch armv8-a                        
    .fpu    neon-vfpv4

// Number of loop iterations for timing purposes:

#define iterations  10000000

    .data
title:  
    .asciz      "Listing9-5"
      
// Buffer to hold the data from u32ToBuf:
  
    .align      2
outBuf: 
    .space      256, 0        

// Values used to test the u32ToBuf/u32ToStr functions:
        
    .align  2  
u32Values:
    .word   0, 1, 2, 3, 4, 5, 6, 7, 8, 9
    .word   10, 11, 15, 16, 20, 30
    .word   40, 50, 60, 64, 64, 99, 100, 101
    .word   127, 128, 200, 255, 256, 300, 400, 500
    .word   511, 512, 600, 700, 800, 900, 999
    .word   1000, 1001, 1023, 1024, 2000, 2047, 2048, 3000, 4095
    .word   4096, 5000, 6000, 7000, 8000, 8191, 8192, 9000, 9999
    .word   10000, 10001, 11000, 12000, 13000, 14000, 15000
    .word   16000, 16383, 16384, 17000, 18000, 19000, 20000 
    .word   30000, 32767, 32768, 40000, 50000, 60000, 65535
    .word   65536, 70000, 80000, 90000, 99999, 100000, 100001
    .word   131071, 131072, 200000, 262143, 262144, 300000
    .word   400000, 500000, 524287, 524288, 600000, 700000
    .word   800000, 900000, 999999, 1000000, 1000001, 1048575
    .word   1048576, 2000000, 2097151, 2097152, 3000000, 4000000
    .word   4194303, 4194304, 5000000, 8388607, 8388608, 9000000
    .word   9999999, 10000000, 10000001, 16777215, 16777216
    .word   33554431, 33554432, 40000000, 50000000, 60000000
    .word   67108863, 67108864, 70000000, 80000000, 90000000
    .word   99999999, 100000000, 100000001, 134217727, 134217728
    .word   200000000, 268435455, 268435456, 300000000, 400000000
    .word   500000000, 536870911, 536870912, 600000000, 700000000
    .word   800000000, 900000000, 999999999, 1000000000
    .word   1000000001, 1073741823, 1073741824, 2000000000
    .word   2147483647, 2147483648, 3000000000
    .word   4000000000, 0xffffffff
    .set    numU32, (.-u32Values)/4
            
        .text
        .align  2
        
        #include "print.inc"

////////////////////////////////////////////////
//
// getTitle
//
//  Returns pointer to program name in R0:
        
        proc    getTitle, thumb, public
        ldr     r0, =title
        bx      lr
        endp    getTitle

////////////////////////////////////////////////
//
//  u32ToBuf
//
//  Converts an integer to string of decimal
// digits using the minimum number of output
// positions.
//
// Inputs:
//  R0: pointer to the buffer.
//  R1: number to convert.
//
// Output:
//  R0: points at the first byte after
//      the string of converted digits.
//
// This code does not preserve any registers!
// Caller is responsible for saving R0-R5 (or
// call u32ToStr to autmatically save them).
    
    proc    u32ToBuf, thumb, public
    
    .macro  rtn
    bx      lr
    .endm       

// Semi-binary search to quickly determine the
// number of output digits and transfer to code
// that handles a specific number of digits:
     
    ldr         r2, =10000
    cmp         r1, r2
    bhs         fivePlus
    
    cmp         r1, #100
    bhs         threeOrFourDigits
    
    cmp         r1, #10
    bhs         twoDigits
    
    // Special case for a single digit:
    
    orr         r2, r1, #'0'
    strb        r2, [r0], #1
    rtn
    
// If the value is less than 100 (but greater than
// 10), then fetch two characters from twoDigitTbl:

twoDigits:
    ldr         r3, =twoDigitTbl
    add         r3, r3, r1, lsl #1
    ldrh        r2, [r3]
    strh        r2, [r0], #2
    rtn

// Handle values in the range 100 to 9999 here:
    
threeOrFourDigits:
    mov         r2, #100
    udiv        r3, r1, r2
    mls         r4, r3, r2, r1  // R1 % 100 to R4.

    cmp         r1, #1000
    bhs         fourDigits

    // Has three digits. Compute R1 / 100 and R1 % 100,
    // then emit the appropriate characters:

    orr         r3, r3, #'0'    // Convert MSD to ASCII.
    strb        r3, [r0], #1
    
    ldr         r3, =twoDigitTbl    // Convert two LO
    add         r3, r3, r4, lsl #1  // digits to ASCII.
    ldrh        r2, [r3]
    strh        r2, [r0], #2
    rtn

// Has four digits, so output two pairs of characters:

fourDigits:
    ldr         r2, =twoDigitTbl
    add         r2, r2, r3, lsl #1  // Most significant
    ldrh        r3, [r2]            // pair of digits.
    strh        r3, [r0], #2

    ldr         r2, =twoDigitTbl
    add         r2, r2, r4, lsl #1  // Least significant
    ldrh        r3, [r2]            // pair of digits.
    strh        r3, [r0], #2
    rtn

// Handle values in the range 10,000 to 1,000,000,000 here:

fivePlus:
    ldr         r2, =1000000
    cmp         r1, r2
    bhs         millionPlus
    
    ldr         r2, =100000
    cmp         r1, r2
    bhs         sixDigits
    
    // Must be five digits, here.
    
fiveDigits:
    ldr         r5, =twoDigitTbl
    mov         r2, #10000
    udiv        r3, r1, r2      // R1 / 10000 to R3.
    mls         r4, r3, r2, r1  // R1 % 10000 to R4.
    
    orr         r2, r3, #'0'    // Output MSD.
    strb        r2, [r0], #1
    
    mov         r2, #100
    udiv        r1, r4, r2      // R4 / 100 to R1.
    mls         r3, r1, r2, r4  // R4 % 100 to R3.
    
    // Output middle two digits:
    
    ldrh        r4, [r5, r1, lsl #1]    
    strh        r4, [r0], #2

    // Output least sigificant digits:
    
    ldrh        r3, [r5, r3, lsl #1]
    strh        r3, [r0], #2
    rtn

// Handle six-digit values here:
    
sixDigits:
    ldr         r5, =twoDigitTbl
    ldr         r2, =10000
    udiv        r3, r1, r2      // R1 / 1000000 to R3
    mls         r4, r3, r2, r1  // R1 % 1000000 to R4

    // Output two HO digits:
    
    ldrh        r3, [r5, r3, lsl #1]
    strh        r3, [r0], #2

    mov         r2, #100
    udiv        r3, r4, r2      // R4 / 1000000 to R3
    mls         r4, r3, r2, r4  // R4 % 1000000 to R4

    // Output middle digits:
    
    ldrh        r3, [r5, r3, lsl #1]
    strh        r3, [r0], #2

    // Output LO digits:
    
    ldrh        r3, [r5, r4, lsl #1]
    strh        r3, [r0], #2
    rtn

// Handle values in the range 10,000,000 to 2,147,483,647:

millionPlus:            
    ldr         r2, =100000000
    cmp         r1, r2
    bhs         nineOrTenDigits

    
    ldr         r2, =10000000
    cmp         r1, r2
    bhs         eightDigits
    
    // Seven output digits here:
    
    ldr         r5, =twoDigitTbl
    ldr         r2, =1000000
    udiv        r3, r1, r2      // R1 / 1000000 to R3
    mls         r4, r3, r2, r1  // R1 % 1000000 to R4
    
    orr         r2, r3, #'0'    // Output MSD
    strb        r2, [r0], #1
    
    mov         r2, #10000
    udiv        r1, r4, r2      // R4 / 10000 to R1
    mls         r3, r1, r2, r4  // R4 % 10000 to R3
    
    // Output the fifth and sixth digits:
    
    ldrh        r4, [r5, r1, lsl #1]
    strh        r4, [r0], #2
    
    mov         r2, #100
    udiv        r4, r3, r2      // R3 / 100 to R4
    mls         r3, r4, r2, r3  // R3 % 100 to R3
    
    // Output the third and fourth digits:
    
    ldrh        r1, [r5, r4, lsl #1]
    strh        r1, [r0], #2
    
    // Output the two LO digits:
    
    ldrh        r1, [r5, r3, lsl #1]
    strh        r1, [r0], #2
    rtn

// Eight output digits here:
    
eightDigits:
    ldr         r5, =twoDigitTbl
    ldr         r2, =1000000
    udiv        r4, r1, r2      // R1 / 10000 to R4
    mls         r3, r4, r2, r1  // R1 % 10000 to R3
    
    // Output the HO two digits:
    
    ldrh        r4, [r5, r4, lsl #1]
    strh        r4, [r0], #2
    
    mov         r2, #10000
    udiv        r4, r3, r2      // R3 / 10000 to R4
    mls         r3, r4, r2, r3  // R3 % 10000 to R3

    // Output the fifth and sixth digits:
    
    ldrh        r1, [r5, r4, lsl #1]
    strh        r1, [r0], #2
    
    mov         r2, #100
    udiv        r4, r3, r2      // R3 / 100 to R4
    mls         r3, r4, r2, r3  // R3 % 100 to R3 
    
    // Output the third and fourth digits:
    
    ldrh        r1, [r5, r4, lsl #1]
    strh        r1, [r0], #2
    
    // Output the LO digits:
    
    ldrh        r1, [r5, r3, lsl #1]
    strh        r1, [r0], #2
    rtn

// Nine or ten output digits here:

nineOrTenDigits:    
    ldr         r2, =1000000000
    cmp         r1, r2
    bhs         tenDigits
    
    // Nine output digits here:
    
    ldr         r5, =twoDigitTbl
    ldr         r2, =100000000
    udiv        r3, r1, r2      // R1 / 100000000 to R3
    mls         r4, r3, r2, r1  // R1 % 100000000 to R4
    
    // Output the HO digit:
    
    orr         r2, r3, #'0'    // Output MSD
    strb        r2, [r0], #1
    
    ldr         r2, =1000000
    udiv        r3, r4, r2      // R4 / 1000000 to R3
    mls         r4, r3, r2, r4  // R4 % 1000000 to R4
    
    // Output the seventh and eighth digits:
    
    ldrh        r4, [r5, r3, lsl #1]
    strh        r4, [r0], #2
    
    mov         r2, #10000
    udiv        r3, r4, r2      // R4 / 10000 to R3
    mls         r4, r3, r2, r4  // R4 % 10000 to R4

    // Output the fifth and sixth digits:
    
    ldrh        r1, [r5, r3, lsl #1]
    strh        r1, [r0], #2
    
    mov         r2, #100
    udiv        r3, r4, r2      // R4 / 100 to R3
    mls         r4, r3, r2, r4  // R4 % 100 to R4 
    
    // Output the third and fourth digits:
    
    ldrh        r1, [r5, r3, lsl #1]
    strh        r1, [r0], #2
    
    // Output the LO digits:
    
    ldrh        r1, [r5, r4, lsl #1]
    strh        r1, [r0], #2
    rtn

// Handle 10 digit output here:
    
tenDigits:
    ldr         r5, =twoDigitTbl
    ldr         r2, =100000000
    udiv        r4, r1, r2      // R1 / 100000000 to R4
    mls         r3, r4, r2, r1  // R1 % 100000000 to R3
    
    // Output the HO two digits:
    
    ldrh        r4, [r5, r4, lsl #1]
    strh        r4, [r0], #2
    
    ldr         r2, =1000000
    udiv        r4, r3, r2      // R3 / 1000000 to R4
    mls         r3, r4, r2, r3  // R3 % 1000000 to R3

    // Output the seventh and eighth digits:

    ldrh        r1, [r5, r4, lsl #1]
    strh        r1, [r0], #2
    
    mov         r2, #10000
    udiv        r4, r3, r2      // R3 / 10000 to R4
    mls         r3, r4, r2, r3  // R3 % 10000 to R3 

    // Output the fifth and sixth digits:
    
    ldrh        r1, [r5, r4, lsl #1]
    strh        r1, [r0], #2
    
    mov         r2, #100
    udiv        r4, r3, r2      // R3 / 100 to R4
    mls         r3, r4, r2, r3  // R3 % 100 to R3 
    
    // Output the third and fourth digits:
    
    ldrh        r1, [r5, r4, lsl #1]
    strh        r1, [r0], #2
    
    // Output the LO digits:
    
    ldrh        r1, [r5, r3, lsl #1]
    strh        r1, [r0], #2
    rtn
    endp    u32ToBuf
    
// Table to convert an integer in the range 0..99 to the corresponding
// pair of ASCII characters. Note that this array contains 100 elements,
// each two bytes long.

    .align  2
twoDigitTbl:
    .ascii "00010203040506070809"
    .ascii "10111213141516171819"
    .ascii "20212223242526272829"
    .ascii "30313233343536373839"
    .ascii "40414243444546474849"
    .ascii "50515253545556575859"
    .ascii "60616263646566676869"
    .ascii "70717273747576777879"
    .ascii "80818283848586878889"
    .ascii "90919293949596979899" 
 
////////////////////////////////////////////////////
//
// u32ToStr
//
// Facade code for u32ToBuf that preserves the
// registers:

    proc    u32ToStr, thumb, public
    push    {r0-r5, lr}
    bl      u32ToBuf
    pop     {r0-r5, pc}
    endp    u32ToStr
    
////////////////////////////////////////////////////
//
// Here's the assembly language main program:
        
        proc    asmMain, thumb, public
        push    {r0-r12, lr}
        
// This code is for timing purposes:

        ldr     r11, =iterations
tstLp:
        ldr     r8, =u32Values
        mov     r9, #numU32
u32DLp:
        ldr     r0, =outBuf
        ldr     r1, [r8], #4
        bl      u32ToStr
        
        subs    r9, r9, #1
        bne     u32DLp
        
        subs    r11, r11, #1
        bne     tstLp

// This code is for testing purposes:

        ldr     r8, =u32Values
        mov     r9, #numU32
prtLp:
        ldr     r0, =outBuf
        ldr     r1, [r8]
        bl      u32ToBuf
        
        mov     r1, #0
        strb    r1, [r0]    // Zero-terminate string.
        
        ldr     r0, [r8], #4
        ldr     r1, =outBuf
        bl      print
        wastr   "testing: %10lu = '%s'\n"       
        
        subs    r9, r9, #1
        bne     prtLp
                
        bl      print
        wastr   "Timing complete\n"
		
        pop     {r0-r12, pc}
        endp    asmMain
